@@ -327,7 +327,7 @@ void exec_pipelines_async(const pipeline_T *p)
     
     if (cpid > 0) {
 	/* parent process: add a new job */
-	job_T *job = xmalloc(sizeof *job + sizeof *job->j_procs);
+	job_T *job = xmalloc(add(sizeof *job, sizeof *job->j_procs));
 	process_T *ps = job->j_procs;
 
 	ps->pr_pid = cpid;
@@ -768,6 +768,11 @@ pid_t exec_process(
 	finally_exit = true;
     }
 
+    lastcmdsubstatus = Exit_SUCCESS;
+
+    /* connect pipes and close leftovers */
+    connect_pipes(pi);
+
     if (c->c_type == CT_SIMPLE) {
 	if (!expand_line(c->c_words, &argc, &argv)) {
 	    laststatus = Exit_EXPERROR;
@@ -787,11 +792,6 @@ pid_t exec_process(
     }
     /* `argc' and `argv' are used only for `CT_SIMPLE'. */
 
-    lastcmdsubstatus = Exit_SUCCESS;
-
-    /* connect pipes and close leftovers */
-    connect_pipes(pi);
-
     /* open redirections */
     savefd_T *savefd;
     if (!open_redirections(c->c_redirs, &savefd)) {
@@ -1941,7 +1941,7 @@ int dot_builtin(int argc, void **argv)
 
 error:
     free(mbsfilename);
-    if (!is_interactive_now)
+    if (special_builtin_executed && !is_interactive_now)
 	exit_shell_with_status(Exit_FAILURE);
     return Exit_FAILURE;
 }
@@ -2074,7 +2074,7 @@ int exec_builtin_2(int argc, void **argv, const wchar_t *as, bool clear)
     }
 
     exec_external_program(commandpath, argc, mbsargv0, argv, envs);
-    err = Exit_NOEXEC;
+    err = laststatus;
 
     if (clear)
 	plfree((void **) envs, free);
